/********************************************************************************************
 *  Ebbie-MM Version 4.3
 *  Copyright (c)  2008 Nov.  H. Alexander Ebhardt, Herbert H. Tsang and Denny Chen Dai
 *  Permission is granted to copy, distribute and/or modify this document
 *  under the terms of the GNU Free Documentation License, Version 1.2
 *  or any later version published by the Free Software Foundation;
 *  with no Invariant Sections, no Front-Cover Texts, and no Back-Cover
 *  Texts.  A copy of the license is included in the section entitled "GNU
 *  Free Documentation License".
 *********************************************************************************************/

package ebbieMM;

import ebbieMM.align.Alignment;
import ebbieMM.fasta.Fasta;
import ebbieMM.fasta.readFasta;
import ebbieMM.util.CharNode;
import ebbieMM.util.RString;
import ebbieMM.util.ResultLog;
import java.io.File;
import java.io.IOException;
import java.util.Vector;

/*
 * Ebbie-MM Main Class
 * 
 * An RNA Sequence Matching Analysis Program
 * This File contains all major Ebbie-MM analysis functions
 * see main() function for the program main loop
 *
 */
public class Main {

    private static String version = "4.3";
    private static ResultLog out = null; /*storing statistics log result*/
    private static ResultLog algout = null; /*store only sequence information in FASTA format*/
 /* private static ResultLog mismatchout = null; store 1-position mismatch for every query (and every mutation)*/


    /** Creates a new instance of Main */
    public Main() {
    }

    /**
    
     * @param args the command line arguments
    
     *
    
     */
    public static void main(String[] args) {

        String os = System.getProperty("os.name");
        String queryF = "", dbF = "", logF = "";
        System.out.println("Ebbie-MM Version \n" + version);
        System.out.println("Running under " + os);

        /*statistics variable*/
        long nFullMatch = 0; /*number of query sequence that is full match*/
        long nOneMis = 0; /* number of one-mismatch leading to full match*/
        long nOther = 0; /* others (more than one mismatch, null blast output)*/
        Vector mutLog = new Vector();/* mutation loggging   A->C(10) */
        boolean setPenalty = false; float penalty = 0;

        /*determine OS type*/
        if (os.startsWith("Windows")) {
            queryF = "blast\\qr\\qr.txt";
            dbF = "blast\\db\\genomeIV.fcgi";
        } else {
            queryF = "blast/qr/qr.txt";
            dbF = "blast/db/genomeIV.fcgi";
        }

        /*help information*/
        if (args.length == 0) {
            System.out.println("USE:  java -jar Ebbie-MM.jar -f [query_file] " + " -db [database_file] " + " [-o] [result_log]" + " -q [penalty]");
            if (os.startsWith("Windows")) {
                System.out.println("\nExample: java -jar Ebbie-MM.jar -f blast\\qr\\qr.txt -db blast\\db\\genomeIV.fcgi -o result.log -q -1");
            } else {
                System.out.println("\nExample: java -jar Ebbie-MM.jar -f blast/qr/qr.txt -db blast/db/genomeIV.fcgi -o result.log -q -1");
            }
            return;
        }

        /*argument processing*/
        for (int i = 0; i < args.length; i++) {
            if (args[i].equals("-f")) {
                queryF = args[++i];
            } else if (args[i].equals("-db")) {
                dbF = args[++i];
            } else if (args[i].equals("-o")) {
                logF = args[++i];
            } else if (args[i].equals("-q")) {
                setPenalty = true;
                penalty = Float.parseFloat(args[++i]);
            }
            else {
                System.out.println("Unrecognized parameters : " + args[i]);
                System.out.println("-f [query_file] " + " -db [database_file] " + " [-o] [result_log]" + " -q [penalty]");
                return;
            }
        }

        /*LOG FILES*/
        out = new ResultLog(logF); /*output log file*/
        algout = new ResultLog(logF + ".fasta"); /*fasta log file*/
      /* mismatchout = new ResultLog(logF + ".mis"); mismatch log file*/
        Vector query_list = (new readFasta()).read(queryF); /*break down query into list of fasta sequences*/
        if (query_list == null) {
            return;
        }

        preAnalysis(query_list); /* pre query_list analysis */
        filterNonStd(query_list); /* filter out sequence containing non-{acgt} symbol */
        filterIdentical(query_list); /* filter out identical small RNA sequence */

        /*initialize Blast Spawner*/
        BlastAlgs algs = new BlastAlgs(dbF,setPenalty, penalty);
        /*Program Main Loop Goes Here*/
        for (int i = 0; i < query_list.size(); i++) {
            Fasta q = (Fasta) query_list.get(i);
          /*  out.println("\n\n\n######## Processing Query Sequence " + i + " ##############\n" + " Identical Count: " + q.count + "\n Sequence Head (all)" + q.head); */
          /*  System.out.println("\n\n\n######## Processing Query Sequence " + i + " ##############\n" + " Identical Count: " + q.count + "\n Sequence Head (all)" + q.head); */
            algout.println(q.head);/*log the header in fasta log file*/
          /*  mismatchout.println("\n\n\n######## Query Sequence " + i + " ##############\n" + "Sequence Head (all)" + q.head);*/

            /* blast the sequence and mutate until full match found */
            int type = mutation(algs, q, mutLog);
            switch (type) {
                case 0:
                    nFullMatch++;
                    break;
                case 1:
                    nOneMis++;
                    break;
                case 2:
                    nOther++;
                    break;
            }
            algout.println("\n\n");
            ram_status();
        }//for
        postAnalysis(nFullMatch, nOneMis, nOther, mutLog);
        algs.terminate();
        System.out.println("EbbieMM exiting....");
    }

    
    
    /**
     * Ebbie-MM Mutation function   
     * Updated: 2008/1/03 
     * 
     * @param algs The BlastAll algorithm
     * @param query The query sequence in Fasta format 
     * @param mutLog the vetoro storing mutation log information
     * 
     * @return:  0: fullmatch  1: 1-mismatch  2: others (more than 1-mismatch,null)
     */
    public static int mutation(BlastAlgs algs, Fasta query, Vector mutLog) {

        int type = 2; /* return type, defaulted to be 2 */
        int count = 0; /* counter for mutation */
        Alignment refAlign = null; /*the referenece alignment to be mutated and kept track of*/
        String log = "";

        do {
            /*blast query sequence*/
            String blastOut = algs.blast_query(query);
            /*parse in the blast result*/
            Vector alignmentSet = parseBlastResultXML(blastOut);
            //Vector alignmentSet = parseBlastResult(blastOut);
            Alignment align = null; /* store the targeted alignment */
            
            /* nothing found */
            if (alignmentSet == null) {
                /* out.println("*** no alignments found ***"); */
                type = 2;
                if (count == 0) /*if first time run and return nothing, then quit*/ {
                    break;
                }
                //otherwise the mutation continues since we did obtain 1-pos 
                //mismatch in previous runs
            }
            
            /* filter out non-ACGTU alignment in blast output */
            if (alignmentSet != null)
                filterBlastResult(alignmentSet);
            
            /* retrieve full match or one-pos mismatch */
            /* first run of blast */
            if (count == 0) 
                refAlign = align = blastH(alignmentSet, query.sequence.length(), null);
            /* following run of blast */
            else
                align = blastH(alignmentSet, query.sequence.length(), refAlign);
                  
            /*log all full & 1-pos mismatch*/
            logBlastMismatch(alignmentSet, query.sequence.length());

            /* nothing found */
            if (align == null) {
               /*out.println("*** no full match or 1-pos mismatch found ***");*/
                if (count == 0) /* if first time run, then out of mutation */ {
                    break;
                }
                //else continue to next mutation blast
                //since previously we obtained a one-position mismatch
            }
            /*either full or 1-pos found*/
            else {
                int misMatchNum = align.checkFullMatch(query.sequence.length());
                /*** full match , break the while loop ***/
                if (misMatchNum == 0) {
                   /* out.println("*** full match found ***"); */
                   /* align.displayAlignment(out); */
                    align.displayAlignment(algout);
                    algout.println("");
                    if (count == 0) {
                        type = 0; /*type 0 since it is first time blastn found full match*/
                    } else {
                        type = 1; /*type 1 (1-mismatch leading to full match)*/
                    }
                    if (!log.equals("")) {
                        mutLog.add(log); /*log the mutation info*/
                    }
                    break; //break the while loop
                } 
                /*** one position mismatch ***/
                else if (misMatchNum == 1) {
                    /* out.println("*** 1-pos mismatch found ***"); */
                    /* align.displayAlignment(out); */
                    align.displayAlignment(algout);
                    algout.println("");
                }
                /*** neighter full match nor 1-pos mismatch ***/
                /*do nothing, go ahead to next mutation then*/
            }//else - either full / one-pos mismatch found 

            log += refAlign.mutate(refAlign.mis_pos);  /* mutate on the query sequnece, log the mutation string */
            query.sequence = refAlign.getQbase(); /* update the mutated query sequence */
            count++; /*update mutation counter*/
            /*out.println("\n## mutation time " + count + " ##"); */
            /*mismatchout.println("\n## mutation time " + count + " ##\n");*/
            /*refAlign.displayMutation(out);*/
        } while (count <= 3);
        
        return type;

    }

    /**
     * Parse XML version of the blast output in string format
     * @param blastResult - Blast output result in xml format
     * @return List of alignment objects 
     */
    public static Vector parseBlastResultXML(String blastResult) {

        Vector alignmentSet = new Vector();
        /*parse xml string*/

        int pos = 0;

        do {

            /*obtain next alignment block*/
            int start = blastResult.indexOf("<Hsp>", pos);
            int end = blastResult.indexOf("</Hsp>", start) + 6;
            /*use alignment block to create alignment*/
            if (start != -1 && end != -1) {
                Alignment elem = new Alignment();
                elem.parseXML(blastResult.substring(start, end));
                alignmentSet.add(elem);
            } else {
                break;
            }
            pos = end;
        } while (true);
        return alignmentSet;
    }

    /*    
     * the func that parse through standard blast output file and construct alignment list      
     *   
     */
    public static Vector parseBlastResult(String blastResult) {
        Vector alignmentSet = new Vector();
        RString data = new RString(blastResult + "@");

        String word = "";

        do {

            word = data.GetNextWord();

        } while (!(word.equals("Score") && data.SeeNextToken().equals("=")) && !word.equals("@"));



        if (word.equals("@")) {
            return null;
        }
        do {

            data.GetNextToken(); //get "=""

            Alignment elem = new Alignment();

            word = elem.parse(data);

            alignmentSet.add(elem);

        } while (word.equals("Score") && data.SeeNextToken().equals("="));

        //end of the alignment list 



        return alignmentSet;

    }

    /**   
     * Log all blast one-mismatch alignment into mismatchLog    
     *    
     * @alignmentSet  the data set containing all blasted alignments    
     * @query_length the length of the original query    
     */
    public static void logBlastMismatch(Vector alignmentSet, int query_length) {
        if (alignmentSet == null) {
            return;
        }
        for (int i = 0; i < alignmentSet.size(); i++) {
            Alignment align = (Alignment) alignmentSet.get(i);
            int misMatchNum = align.checkFullMatch(query_length);
            
        }
    }

    /* Retrieve top-scoring fullmatch or 1-pos mismatch from blast output    
     *    
     * refAlign - the reference alignment we try to keep track of (if not NULL)   
     * alignmentSet - the alignment output from blast  
     * query_length - the original length of the query   
     *  
     * if refAlign != null, we use this to recover the alignment from blast (return null if lose track)  
     */
    public static Alignment blastH(Vector alignmentSet, int query_length, Alignment refAlign) {

        Alignment obj = null;
        if (alignmentSet == null) {
            return null;
        }
        for (int i = 0; i < alignmentSet.size(); i++) {
            Alignment align = (Alignment) alignmentSet.get(i);
            int misMatchNum = align.checkFullMatch(query_length);
            if (misMatchNum == 1 || misMatchNum == 0) /* either full / 1-pos mismatch */ {
                if (refAlign != null) /* retrieve reference alignment */ {
                    if (refAlign.sbjctHead == align.sbjctHead &&
                            refAlign.sbjctTail == align.sbjctTail) 
                        /*check if the aligned segment is the same as reference alignment*/ {
                        obj = align;
                        break;
                    }
                } else /*simply retrieve the very first full/1-mis found*/ {
                    obj = align;
                    break;
                }
            }
        }
        return obj;
    }

    /* 
    
     * filter out alignment whose query sequence contains non-ACGT symbol
    
     * but also database sequence containing non-ACGT symbol
    
     */
    public static void filterBlastResult(Vector alignmentSet) {

        for (int i = 0; i < alignmentSet.size(); i++) {

            Alignment align = (Alignment) alignmentSet.get(i);

            for (int j = 0; j < align.qbase.size(); j++) {

                String tmp = align.qbase.get(j).toString().toUpperCase();

                String tmp2 = align.sbase.get(j).toString().toUpperCase();

                if (!(tmp.equals("A") || tmp.equals("C") || tmp.equals("G") || tmp.equals("T") || tmp.equals("U")) ||
                        !(tmp2.equals("A") || tmp2.equals("C") || tmp2.equals("G") || tmp2.equals("T") || tmp2.equals("U"))) {

                    /*out.println("\n##Filtering out non-ACGT alignment\n");*/

                    alignmentSet.remove(i);

                    i--;

                    break;

                }

            }

        }



    }

    /* 
    
     * report if and only if there are multiple 1-mismatch having the same highest score
    
     * @return the number of 1-mismatch having the same highest score 
    
     * report 0 if fullmatch(or no 1-pos mismatch), 1 if only one high score, others indicate multiple
    
     */
    public static int reportMultipleOptimal(Vector alignmentSet, int query_length) {

        int opt_count = 0; /*count of multiple 1-mismatch optimality*/

        float opt_score = -1;

        Alignment obj = null;

        for (int i = 0; i < alignmentSet.size(); i++) {

            Alignment align = (Alignment) alignmentSet.get(i);

            if (align.qbase.size() == query_length) {  /*a full-length report (full_match or 1-mismatch)*/

                if (align.checkFullMatch(query_length) == 1) { //1-pos mismatch 

                    if (opt_score == -1) {   //very first highest score 

                        opt_score = align.score; //first opt

                        opt_count++;

                    } else {  // somewhere in the alignment list 

                        if (align.score == opt_score) //another 1-pos mismatch having same opt score
                        {
                            opt_count++;
                        } else if (align.score < opt_score) {
                            break; // now down to scores lower than opt, no need for further check

                        }
                    }

                }

            }

        }

        return opt_count;

    }

    /*
    
     * execute a system call to blast program 
    
     *
    
     *@blastCmd the blast call command
    
     *@queryF  the small RNA sequence query file
    
     *@dbF  sequence database file 
    
     *@blastOutF   blast output file 
    
     *@return true upon successful execution, otherwise false
    
     */
    public static boolean exeBlastAll(String blastCmd, String queryF, String dbF, String blastOutF) {



        boolean succeed = false;

        Process p = null;

        do {

            /*execute blast*/

            try {

                p = Runtime.getRuntime().exec( //"cmd /c " +

                        blastCmd +
                        " -d " + dbF +
                        " -i " + queryF +
                        " -o " + blastOutF,
                        null, //env added

                        null //run under this dir 
                        );

            } catch (IOException e) {

                System.err.println(e.getMessage());

                out.println(e.getMessage());

            }

            /*check system call status*/

            if (p == null) {

                System.err.println("*** failed to execute BlastN, wait and try again ***");

                out.println("*** failed to execute BlastN, wait and try again ***");

                try {
                    Thread.sleep(1000);
                } catch (InterruptedException e) {
                }

            } else {
                succeed = true;
            }
        } while (!succeed);



        /*wait for blast to finish execution*/

        try {

            p.waitFor();

        } catch (InterruptedException e) {

            Thread.currentThread().interrupt();

        }



        /* out.println(" blasted......\n output_file: " + blastOutF); */

        return true;

    }

    /* 
    
     * OBSOLATED FUNCTION  
    
     * Pre analysis of the query list (list of small RNA in FASTA format)
    
     * A summary page, and identical analysis 
    
     */
    public static void preAnalysis(Vector query_list) {


        out.println("\nEbbie-(mis)match version 4.3");
        out.println("\nReference: Ebhardt HA et al.: XX TBD XX");
        out.println("\nURL: http://www.bioinformatics.org/ebbie/");
        

        out.println("\n################PRE QUERY ANALYSIS#########################\n");

        out.println(" number of query sequence: " + query_list.size());

        out.println("\n################END OF ANALYSIS#########################\n");

    }

    public static void postAnalysis(long nFullMatch, long nOneMis, long nOther, Vector mutLog) {



        out.println("\n################POST QUERY ANALYSIS#########################\n");



        out.println("## number of full match found: " + nFullMatch);

        out.println("## number of one mismatch (leading to full match after mutation): " + nOneMis);



        int[] mut = new int[12];

        // A->C  A-G A-T

        // C->A  C-G C-T

        //G->A   G->C  G->T

        //T->A T->C T->G

        for (int i = 0; i < mutLog.size(); i++) {

            //System.out.println(mutLog.get(i).toString());

            String tmp = (String) mutLog.get(i);

            char f = 'N';

            char e = 'N';

            for (int j = tmp.length() - 1; j >= 0; j--) {

                if (tmp.charAt(j) == '(') {

                    e = tmp.charAt(j - 1);

                    break;

                }//if

            }//for

            f = tmp.charAt(0);



            if (f == 'A' && e == 'C') {
                mut[0]++;
            } else if (f == 'A' && e == 'G') {
                mut[1]++;
            } else if (f == 'A' && e == 'T') {
                mut[2]++;
            } else if (f == 'C' && e == 'A') {
                mut[3]++;
            } else if (f == 'C' && e == 'G') {
                mut[4]++;
            } else if (f == 'C' && e == 'T') {
                mut[5]++;
            } else if (f == 'G' && e == 'A') {
                mut[6]++;
            } else if (f == 'G' && e == 'C') {
                mut[7]++;
            } else if (f == 'G' && e == 'T') {
                mut[8]++;
            } else if (f == 'T' && e == 'A') {
                mut[9]++;
            } else if (f == 'T' && e == 'C') {
                mut[10]++;
            } else if (f == 'T' && e == 'G') {
                mut[11]++;
            }
        }//for

        out.println( mut[3] + " A-to-C ");
        out.println( mut[6] + " A-to-G ");
        out.println( mut[9] + " A-to-U ");
        out.println( mut[0] + " C-to-A ");
        out.println( mut[7] + " C-to-G ");
        out.println( mut[10] + " C-to-U ");
        out.println( mut[1] + " G-to-A ");
        out.println( mut[4] + " G-to-C ");
        out.println( mut[11] + " G-to-U ");
        out.println( mut[2] + " T-to-A ");
        out.println( mut[5] + " T-to-C ");
        out.println( mut[8] + " T-to-G ");


        out.println("## number of others : " + nOther);



        out.println("\n################END OF ANALYSIS#########################\n");

    }

    /*
    
     * Filter out query sequences containing non ACGTU symbol
    
     *
    
     *@query_list  the data set containing the query sequences
    
     */
    public static void filterNonStd(Vector query_list) {

        out.println("\n###############Scanning for non-ACGT sequence##############\n");

        for (int i = 0; i < query_list.size(); i++) {

            Fasta query = (Fasta) query_list.get(i);
            if(query.sequence.length()==0){
                System.out.println("warning...empty fasta sequence string...");
                query_list.remove(i);
                i--;
                continue;
            }

            for (int j = 0; j < query.sequence.length(); j++) {
                char tmp = query.sequence.toUpperCase().charAt(j);
                if (!(tmp == 'A' || tmp == 'C' || tmp == 'G' || tmp == 'T' || tmp == 'U')) {
                    out.println("Removing sequence " + query.head + ".....");
                    query_list.remove(i);
                    i--;
                    break;
                }//if
            }//for
        }
        out.println("\n##################Finish Scanning#################\n");

    }

    /*
    
     * Filter out identical query sequence 
    
     * We buid a sequence tree and try to detect identical sequence through one scan of the data set
    
     * via information recorded in the sequence tree. 
    
     * For identical sequence, we remove it from the query list (but maitaining only one)
    
     * but preserve its header information by appending it to the unique sequence  
    
     */
    public static void filterIdentical(Vector query_list) {



        out.println("\n################Filtering the Sequence##############\n");

        int delCount = 0;

        CharNode tree = new CharNode("root");



        for (int i = 0; i < query_list.size(); i++) {

            Fasta query = (Fasta) query_list.get(i);

            CharNode curN = tree; /*begin with root*/



            for (int j = 0; j < query.sequence.length(); j++) {

                /*take each query sequence*/

                char tmp = query.sequence.toUpperCase().charAt(j);

                /*travers & grow the char tree according to the sequence content*/

                if (tmp == 'A') {

                    if (curN.aNode == null) {
                        curN.aNode = new CharNode("A");
                    }
                    curN = curN.aNode;

                } else if (tmp == 'C') {

                    if (curN.cNode == null) {
                        curN.cNode = new CharNode("C");
                    }
                    curN = curN.cNode;

                } else if (tmp == 'G') {

                    if (curN.gNode == null) {
                        curN.gNode = new CharNode("G");
                    }
                    curN = curN.gNode;

                } else if (tmp == 'U' || tmp == 'T') {

                    if (curN.uNode == null) {
                        curN.uNode = new CharNode("U");
                    }
                    curN = curN.uNode;

                }

            }



            /*now we finish traverse the sequence tree, curN points to the leaf node*/

            if (curN.pointer == -1) /*first time encounter this sequence*/ {

                curN.pointer = i; //store this sequence's position in query list

                query.count = 1; //update the counter 

            } else /*now we found another identical sequence*/ {

                Fasta tmpQ = (Fasta) query_list.get(curN.pointer);//locate the recorded unique sequence

                tmpQ.count++;  //update its counter 

                tmpQ.head += query.head; //concatenate the head info to unique sequence

                query_list.remove(i); //now remove this one

                delCount++; //update total_deletion_counter

                i--; //shift back the looper

            }// 

        }//for



        out.println("Total identical entries removed: " + delCount);

        out.println("\n##########Finish Filtering#####################\n");



    }

    private static void ram_status() {

        Runtime r = Runtime.getRuntime();

        // amount of unallocated memory

        long free = r.freeMemory();



        // total amount of memory available allocated and unallocated.

        long total = r.totalMemory();



        // total amount of allocated memory

        long inuse = r.totalMemory() - r.freeMemory();



        // max memory JVM will attempt to use

        long max = r.maxMemory();



        System.out.printf("***Memory usage: %.1f MB ( %.1f MB)***\n", (float) inuse / (float) (1024 * 1024),
                (float) total / (float) (1024 * 1024));

        // suggest now would be a good time for garbage collection



        if ((float) inuse / (float) total > 0.25) {

            System.gc();

            System.out.println("\ncleaning memory -- garbage collection\n");

        }

    }

    /**
    
     * File Deletion Method.  Invoked by main(), and designed
    
     * for use by other programs as well.  It first makes sure that the 
    
     * specified file or directory is deleteable before attempting to delete it.
    
     * If there is a problem, it throws an IllegalArgumentException.
    
     */
    public static void delete(String filename) {

        File f = new File(filename);

        if (!f.exists()) {
            fail("Delete: no such file or directory: " + filename);
        }
        if (!f.canWrite()) {
            fail("Delete: write protected: " + filename);
        }
        if (f.isDirectory()) {

            String[] files = f.list();

            if (files.length > 0) {
                fail("Delete: directory not empty: " + filename);
            }
        }



        // If we passed all the tests, then attempt to delete it

        boolean success = f.delete();

        // And throw an exception if it didn't work for some (unknown) reason.

        // For example, because of a bug with Java 1.1.1 on Linux, 

        // directory deletion always fails 

        if (!success) {
            fail("Delete: deletion failed");
        }
    }

    /** A convenience method to throw an exception */
    protected static void fail(String msg) throws IllegalArgumentException {

        throw new IllegalArgumentException(msg);

    }
}


